/***************************************************************************
                          codeparser.h  -  description
                             -------------------
    begin                : Die Jul 9 2002
    copyright            : (C) 2002 by André Simon
    email                : andre.simon1@gmx.de
 ***************************************************************************/

/***************************************************************************
 *                                                                         *
 *   This program is free software; you can redistribute it and/or modify  *
 *   it under the terms of the GNU General Public License as published by  *
 *   the Free Software Foundation; either version 2 of the License, or     *
 *   (at your option) any later version.                                   *
 *                                                                         *
 ***************************************************************************/

#ifndef CODEPARSER_H
#define CODEPARSER_H

#include <iostream>
#include <sstream>
#include <string>
#include "languagedefinition.h"
#include "documentstyle.h"
#include "ASFormatter.h"
#include "linewrapper.h"

/**Klasse mit Parse-Methoden, von der die einzelnen Ausgabeklassen
   (zB HtmlCode, RtfCode) ableiten
  *@author André Simon
  */

#define NUMBER_OF_STYLES 11 

namespace highlight {

enum State {
  STANDARD=0,
  KEYWORD,
  TYPE,
  STRING,
  NUMBER,
  SINGLE_LINE_COMMENT,
  MULTI_LINE_COMMENT_BEGIN,
  ESC_CHAR,
  DIRECTIVE_LINE,
  DIRECTIVE_STRING,
  LINENUMBER,

  // Konstanten ab hier duefen nicht mehr als Array-Indizes benutzt werden!!
  MULTI_LINE_COMMENT_END,
  STRING_END,
  DIRECTIVE_LINE_END,
  TAG_BEGIN,
  TAG_END,
  MEMBER_ATTRIBUTE_BEGIN,
  MEMBER_ATTRIBUTE_END,

  _UNKNOWN=100,
  _EOL,
  _EOF,
  _WS
} ;

enum ParseError{
   PARSE_OK,
   BAD_INPUT,
   BAD_OUTPUT,
   BAD_STYLE
};


class CodeParser
  {

  public:
    CodeParser();

    //! CodeParser Konstruktor
    /** \param langDef language definition which contains parsing information 
               about the programming language
        \param printLineNumbers true if line numbers should be printed
        \param cssStyle Name of coloring style being used
        \param numSpaces number of spaces that replace a tab
        \param lineWrappingStyle 0 = kein Wrapping, 1 Wrapping, 2 Wrapping
               mit Einrckung nach "({=" 
    */
    CodeParser(const highlight::LanguageDefinition &langDef, 
               bool printLineNumbers,
               const string &cssStyle, 
               int numSpaces, 
               int lineWrappingStyle, 
               bool fragment);    

    virtual ~CodeParser();

    void setFormatter( enum FormatStyle style);
    void setLanguageDefinition(const LanguageDefinition &langDef);

    // each Parser has to define how to mask its characters
    virtual string maskCharacter(unsigned char ) = 0;

    // apply maskCharacter to s, return the masked string
    string maskString(const string & ) ;

    /**Style-Tags und Bezeichnungen*/
    string styleTagOpen[NUMBER_OF_STYLES],
           styleTagClose[NUMBER_OF_STYLES];

    /** Zeilenumbruch*/
    string newLineTag;

    /**Leerzeichen*/
    string spacer;

    /**Einleiten/Abschliessen von Whitespace-Maskierung*/
    string maskWsBegin, maskWsEnd;

    // Sprachinformationen
    LanguageDefinition langInfo;

    // Flagvariablen
    bool showLineNumbers;

    //True, falls Whitespaces in processWsState maskiertz werden sollen
    // (zB fuer LaTeX)    
    bool maskWs;
    
    //True, falls Header und Footer weggelassen werden
    //nur interessant fr LateX ausgabe, um Quotes zu �dern
    bool fragmentOutput;
    
    //Name des FarbStils
    string  styleName;

    // Anzahl von Leerzeichen, die ein Tab ersetzen sollen
    int numberSpaces;

    //Zeiger auf den zu benutzenden Ein-/Ausgabestrom
    istream *in;
    ostream *out;

    /** sucht ab Position searchPos Ziffer in s und liefert
    Integerwert der gefundenen Ziffer zurueck. */
    enum State getState(const string &s, unsigned int searchPos);

    /** liefert Identifier ab Stringposition index-1 zurueck. */
    string getIdentifier();

    /** liefert Zahl ab Stringposition index-1 zurueck.*/
    string getNumber();

    /** fuegt am Anfang der neuen Zeile eine fortlaufende Nummer ein
        wenn Nummer nicht 4stellig, wird mit spaceStr rechtsbuendig aufgefuellt */
    virtual void insertLineNumber (/*enum State envokingState*/);

    /** gibt Dateiende aus */
    virtual void printFooter() = 0;

    /** gibt formatierten Sourcecode aus */
    virtual void printBody() = 0;

    /** gibt Dateikopf aus */
    virtual void printHeader(const string &) = 0;

    enum ParseError printOutput (const string&, const string&);

    /** gibt Zeilennummer der n�hsten Zeile zurck */
    unsigned int getLineNumber();
    
    void reset();

    // Zaehler der Zeilennmummern
    unsigned int lineNumber,

    // Zaehler, der aktuelle Position in der geparsten Codezeile enthaelt
    lineIndex;

    // Beschreibung der Document-Formatierung
    DocumentStyle *docStyle;


  protected:
                                     
    // aktuelle Zeile, die formatiert werden soll
    string line,

    // String mit Symbolen der jeweiligen Progarmmiersprache 
    // ( z.B //, /*, */, # etc.; und den zugehoerigen Style-Konstanten)
    symbolString;

    /** Anfangszustand des Parsers */
    void processStandardState();

    bool assignInStream(const string&);
    bool assignOutStream(const string&);
    
    // Zeigt den aktuellen Zustand an
    // wird nicht in getCurrentState gesetzt, da nur Zustände interessant
    // sind, die als Index auf die styleCloseTags und styleOpenTags verwendet
    // werden können
    enum State currentState;

  private:    
    
    /** liest Zeile von InputStream ein*/
    bool readNewLine();
    
    /** liefert den n?hsten Character aus dem Inputstream */
    unsigned char getInputChar();    

    /** liefert den letzten gelesenen Character */
    unsigned char getLastChar();    
    
    /**letztes Zeichen der letzten gelesenen Zeile*/
    unsigned char terminatingChar;

    /** AStyle Reformatter*/
    astyle::ASFormatter *formatter;
    
    LineWrapper *lineWrapper;      

    /** enthalt das aktuell geparste Token*/
    string token;

    /** untersucht token und gibt den Zustand zurueck, den token einleitet */
    enum State getCurrentState ( bool lastStateWasNumber=false);

    /** arbeitet einfache Zustaende ab
     token enth�t das erkannte Token, durch den der Zustand gewechselt wurde
     myState ist der Zustand, in den gewechselt wird*/
    bool processTrivialState(enum State myState) ;

    /** Methoden zur Verarbeitung der einzelnen Zustaende*/
    bool processMultiLineCommentState();
    bool processSingleLineCommentState();
    bool processStringState( enum State oldState);
    bool processEscapeCharState();
    bool processDirectiveState();
    bool processTagState();
    void processWsState();

    /** gibt true zurck, falls c ein erlaubter Character innerhalb von Keyword oder Typbezeichner ist */
    bool isAllowedChar(char c) ;

    /** gibt true zurck, das aktuelle Token das erste in der Zeile und kein Whitespace ist */
    bool isFirstNonWsChar() ;

    /** gibt maskierten Token aus und setzt token auf ""*/
    void printMaskedToken();

    /** gibt Escapesequenz aus*/
    void skipEscapeSequence();

    void closeTag(enum State s);

    void openTag(enum State s);

    /** gibt String zum Zeilenumbruch aus*/  
    virtual string getNewLine();
  };
}

#endif
